"""
Linear quadratic regulator class.
"""

import numpy as np
import scipy as sp

from .ctrlutil import is_stabilizable, make_cost_matrix
from .discretization import discretize_ab


class LinearQuadraticRegulator:
    """Linear-quadratic regulator class."""

    def __init__(self, A, B, Qelems, Relems, dt):
        """
        Constructs a LinearQuadraticRegulator.

        Parameter ``A``:
            Continuous system matrix.
        Parameter ``B``:
            Continuous input matrix.
        Parameter ``Qelems``:
            The maximum desired error tolerance for each state.
        Parameter ``Relems``:
            The maximum desired control effort for each input.
        Parameter ``dt``:
            Discretization timestep.
        """
        self.states = A.shape[0]
        self.inputs = B.shape[1]

        discA, discB = discretize_ab(A, B, dt)

        if not is_stabilizable(discA, discB):
            raise RuntimeError(
                f"The system is unstabilizable!\n\nA = {discA}\nB = {discB}\n"
            )

        Q = make_cost_matrix(Qelems)
        R = make_cost_matrix(Relems)

        S = sp.linalg.solve_discrete_are(a=discA, b=discB, q=Q, r=R)

        # K = (BᵀSB + R)⁻¹BᵀSA
        self.K = np.linalg.solve(discB.T @ S @ discB + R, discB.T @ S @ discA)

        self.r = np.zeros((self.states, 1))
        self.u = np.zeros((self.inputs, 1))

    def reset(self):
        """Resets the controller."""
        self.r = np.zeros((self.states, 1))
        self.u = np.zeros((self.inputs, 1))

    def calculate(self, x, r=None):
        """
        Returns the next output of the controller.

        Parameter ``x``:
            The current state vector x.
        Parameter ``r``:
            The current reference vector r (default: previous reference).
        """
        if r is not None:
            self.r = r
        self.u = self.K @ (self.r - x)
        return self.u

    def latency_compensate(self, A, B, dt, input_delay):
        """
        Adjusts LQR controller gain to compensate for a pure time delay in the
        input.

        Linear-Quadratic regulator controller gains tend to be aggressive. If
        sensor measurements are time-delayed too long, the LQR may be unstable.
        However, if we know the amount of delay, we can compute the control
        based on where the system will be after the time delay.

        See https://file.tavsys.net/control/controls-engineering-in-frc.pdf
        appendix C.4 for a derivation.

        Parameter ``A``:
            Continuous system matrix.
        Parameter ``B``:
            Continuous input matrix.
        Parameter ``dt``:
            Discretization timestep.
        Parameter ``input_delay``:
            Input time delay.
        """
        discA, discB = discretize_ab(A, B, dt)
        self.K = self.K @ sp.linalg.fractional_matrix_power(
            discA - discB @ self.K, input_delay / dt
        )



"""Tests the numerical_jacobian module."""

import numpy as np

from frccontrol.numerical_jacobian import numerical_jacobian_x, numerical_jacobian_u


def test_numerical_jacobian():
    """Tests numerical_jacobian_x and numerical_jacobian_u."""

    A = np.array([[1, 2, 4, 1], [5, 2, 3, 4], [5, 1, 3, 2], [1, 1, 3, 7]])
    B = np.array([[1, 1], [2, 1], [3, 2], [3, 7]])

    def AxBuFn(x, u):
        return A @ x + B @ u

    newA = numerical_jacobian_x(4, 4, AxBuFn, np.zeros((4, 1)), np.zeros((2, 1)))
    assert np.allclose(A, newA)

    newB = numerical_jacobian_u(4, 2, AxBuFn, np.zeros((4, 1)), np.zeros((2, 1)))
    assert np.allclose(B, newB)



"""Functions for generating 1D motion profiles."""

import math


def generate_trapezoid_profile(max_v, time_to_max_v, dt, goal):
    """
    Creates a trapezoid profile with the given constraints.

    Parameter ``max_v``:
        Maximum velocity of profile.
    Parameter ``time_to_max_v``:
        Time from rest to maximum velocity.
    Parameter ``dt``:
        Timestep.
    Parameter ``goal``:
        Final position when the profile is at rest.
    Returns:
        t_rec -- List of timestamps.
        x_rec -- List of positions at each timestep.
        v_rec -- List of velocities at each timestep.
        a_rec -- List of accelerations at each timestep.
    """
    t_rec = [0.0]
    x_rec = [0.0]
    v_rec = [0.0]
    a_rec = [0.0]

    a = max_v / time_to_max_v
    time_at_max_v = goal / max_v - time_to_max_v

    # If profile is short
    if max_v * time_to_max_v > goal:
        time_to_max_v = math.sqrt(goal / a)
        time_from_max_v = time_to_max_v
        time_total = 2.0 * time_to_max_v
        profile_max_v = a * time_to_max_v
    else:
        time_from_max_v = time_to_max_v + time_at_max_v
        time_total = time_from_max_v + time_to_max_v
        profile_max_v = max_v

    while t_rec[-1] < time_total:
        t = t_rec[-1] + dt
        t_rec.append(t)
        if t < time_to_max_v:
            # Accelerate up
            a_rec.append(a)
            v_rec.append(a * t)
        elif t < time_from_max_v:
            # Maintain max velocity
            a_rec.append(0.0)
            v_rec.append(profile_max_v)
        elif t < time_total:
            # Accelerate down
            decel_time = t - time_from_max_v
            a_rec.append(-a)
            v_rec.append(profile_max_v - a * decel_time)
        else:
            a_rec.append(0.0)
            v_rec.append(0.0)
        x_rec.append(x_rec[-1] + v_rec[-1] * dt)
    return t_rec, x_rec, v_rec, a_rec


def generate_s_curve_profile(max_v, max_a, time_to_max_a, dt, goal):
    """
    Returns an s-curve profile with the given constraints.

    Parameter ``max_v``:
        Maximum velocity of profile.
    Parameter ``max_a``:
        Maximum acceleration of profile.
    Parameter ``time_to_max_a``:
        Time from rest to maximum acceleration.
    Parameter ``dt``:
        Timestep.
    Parameter ``goal``:
        Final position when the profile is at rest.
    Returns:
        t_rec -- List of timestamps.
        x_rec -- List of positions at each timestep.
        v_rec -- List of velocities at each timestep.
        a_rec -- List of accelerations at each timestep.
    """
    t_rec = [0.0]
    x_rec = [0.0]
    v_rec = [0.0]
    a_rec = [0.0]

    j = max_a / time_to_max_a
    short_profile = max_v * (time_to_max_a + max_v / max_a) > goal

    if short_profile:
        profile_max_v = max_a * (
            math.sqrt(goal / max_a - 0.75 * time_to_max_a**2) - 0.5 * time_to_max_a
        )
    else:
        profile_max_v = max_v

    # Find times at critical points
    t2 = profile_max_v / max_a
    t3 = t2 + time_to_max_a
    if short_profile:
        t4 = t3
    else:
        t4 = goal / profile_max_v
    t5 = t4 + time_to_max_a
    t6 = t4 + t2
    t7 = t6 + time_to_max_a
    time_total = t7

    while t_rec[-1] < time_total:
        t = t_rec[-1] + dt
        t_rec.append(t)
        if t < time_to_max_a:
            # Ramp up acceleration
            a_rec.append(j * t)
            v_rec.append(0.5 * j * t**2)
        elif t < t2:
            # Increase speed at max acceleration
            a_rec.append(max_a)
            v_rec.append(max_a * (t - 0.5 * time_to_max_a))
        elif t < t3:
            # Ramp down acceleration
            a_rec.append(max_a - j * (t - t2))
            v_rec.append(max_a * (t - 0.5 * time_to_max_a) - 0.5 * j * (t - t2) ** 2)
        elif t < t4:
            # Maintain max velocity
            a_rec.append(0.0)
            v_rec.append(profile_max_v)
        elif t < t5:
            # Ramp down acceleration
            a_rec.append(-j * (t - t4))
            v_rec.append(profile_max_v - 0.5 * j * (t - t4) ** 2)
        elif t < t6:
            # Decrease speed at max acceleration
            a_rec.append(-max_a)
            v_rec.append(max_a * (t2 + t5 - t - 0.5 * time_to_max_a))
        elif t < t7:
            # Ramp up acceleration
            a_rec.append(-max_a + j * (t - t6))
            v_rec.append(
                max_a * (t2 + t5 - t - 0.5 * time_to_max_a) + 0.5 * j * (t - t6) ** 2
            )
        else:
            a_rec.append(0.0)
            v_rec.append(0.0)
        x_rec.append(x_rec[-1] + v_rec[-1] * dt)
    return t_rec, x_rec, v_rec, a_rec



"""
Linear plant inversion feedforward.
"""

import numpy as np

from .discretization import discretize_ab


class LinearPlantInversionFeedforward:
    """A linear plant inversion feedforward."""

    def __init__(self, A, B, dt):
        """
        Constructs a linear plant inversion feedforward.

        Parameter ``A``:
            Continuous system matrix of the plant being controlled.
        Parameter ``B``:
            Continuous input matrix of the plant being controlled.
        Parameter ``dt``:
            Discretization timestep.
        """
        self.A, self.B = discretize_ab(A, B, dt)
        self.r = np.zeros((A.shape[0], 1))
        self.u_ff = np.zeros((B.shape[1], 1))

    def reset(self, initial_reference):
        """
        Resets the feedforward with a specified initial reference vector.

        Parameter ``initial_reference``:
            The initial reference vector.
        """
        self.r = initial_reference
        self.u_ff = np.zeros(self.u_ff.shape)

    def calculate(self, next_r):
        """
        Calculate the feedforward with only the desired future reference. This
        uses the internally stored "current" reference (calling reset() will
        override this).

        Parameter ``next_r``:
            The reference state of the future timestep (k + 1).
        Returns:
            The calculated feedforward.
        """
        self.u_ff = np.linalg.pinv(self.B) @ (next_r - (self.A @ self.r))
        self.r = next_r
        return self.u_ff



"""Control system utility functions."""

import numpy as np
import scipy as sp


def make_cost_matrix(elems):
    """
    Creates a cost matrix from the given vector for use with LQR.

    The cost matrix is constructed using Bryson's rule. The inverse square
    of each element in the input is taken and placed on the cost matrix
    diagonal.

    Parameter ``elems``:
        A vector. For a Q matrix, its elements are the maximum allowed
        excursions of the states from the reference. For an R matrix, its
        elements are the maximum allowed excursions of the control inputs from
        no actuation.
    Returns:
        State excursion or control effort cost matrix.
    """
    return np.diag([0.0 if elem == float("inf") else 1.0 / elem**2 for elem in elems])


def make_cov_matrix(elems):
    """
    Creates a covariance matrix from the given vector for use with Kalman
    filters.

    Each element is squared and placed on the covariance matrix diagonal.

    Parameter ``elems``:
        A vector. For a Q matrix, its elements are the standard deviations of
        each state from how the model behaves. For an R matrix, its elements are
        the standard deviations for each output measurement.
    Returns:
        Process noise or measurement noise covariance matrix.
    """
    return np.diag(np.square(elems))


def is_stabilizable(A, B):
    """
    Returns true if (A, B) is a stabilizable pair.

    (A, B) is stabilizable if and only if the uncontrollable eigenvalues of A,
    if any, have absolute values less than one, where an eigenvalue is
    uncontrollable if rank([λI - A, B]) < n where n is the number of states.

    Parameter ``A``:
        System matrix.
    Parameter ``B``:
        Input matrix.
    """
    rows = A.shape[0]
    w, _ = sp.linalg.eig(A)

    for i in range(rows):
        if abs(w[i]) < 1:
            continue

        E = np.block([[w[i] * np.eye(rows) - A, B]])

        if np.linalg.matrix_rank(E) < rows:
            return False
    return True


def is_detectable(A, C):
    """
    Returns true if (A, C) is a detectable pair.

    (A, C) is detectable if and only if the unobservable eigenvalues of A, if
    any, have absolute values less than one, where an eigenvalue is unobservable
    if rank([λI - A; C]) < n where n is the number of states.

    Parameter ``A``:
        System matrix.
    Parameter ``C``:
        Output matrix.
    """
    return is_stabilizable(A.T, C.T)


def ctrb(A, B):
    """
    Returns the controllability matrix for A and B

    Parameter ``A``:
        System matrix.
    Parameter ``B``:
        Input matrix.
    """
    return np.hstack(
        [B] + [np.linalg.matrix_power(A, i) @ B for i in range(1, A.shape[0])]
    )


def obsv(A, C):
    """
    Returns the observability matrix for A and C

    Parameter ``A``:
        System matrix.
    Parameter ``C``:
        Output matrix.
    """
    return np.vstack(
        [C] + [C @ np.linalg.matrix_power(A, i) for i in range(1, A.shape[0])]
    )



"""Functions for integrating nonlinear system dynamics."""

import math
import numpy as np


def rk4(f, x, u, dt):
    """
    Performs 4th order Runge-Kutta integration of dx/dt = f(x, u) for dt.

    Parameter ``f``:
        Vector function to integrate.
    Parameter ``x``:
        Vector of states.
    Parameter ``u``:
        Vector of inputs (constant for dt).
    Parameter ``dt``:
        Time for which to integrate.
    """
    half_dt = dt * 0.5
    k1 = f(x, u)
    k2 = f(x + half_dt * k1, u)
    k3 = f(x + half_dt * k2, u)
    k4 = f(x + dt * k3, u)
    return x + dt / 6.0 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)


def rkdp(f, x, u, dt, max_error=1e-6):
    """
    Performs adaptive Dormand-Prince integration of dx/dt = f(x, u) for dt.

    Parameter ``f``:
        Vector function to integrate.
    Parameter ``x``:
        Vector of states.
    Parameter ``u``:
        Vector of inputs (constant for dt).
    Parameter ``dt``:
        Time for which to integrate.
    """
    A = np.empty((6, 6))

    A[0, :1] = np.array([1.0 / 5.0])
    A[1, :2] = np.array([3.0 / 40.0, 9.0 / 40.0])
    A[2, :3] = np.array([44.0 / 45.0, -56.0 / 15.0, 32.0 / 9.0])
    A[3, :4] = np.array(
        [19372.0 / 6561.0, -25360.0 / 2187.0, 64448.0 / 6561.0, -212.0 / 729.0]
    )
    A[4, :5] = np.array(
        [
            9017.0 / 3168.0,
            -355.0 / 33.0,
            46732.0 / 5247.0,
            49.0 / 176.0,
            -5103.0 / 18656.0,
        ]
    )
    A[5, :6] = np.array(
        [
            35.0 / 384.0,
            0.0,
            500.0 / 1113.0,
            125.0 / 192.0,
            -2187.0 / 6784.0,
            11.0 / 84.0,
        ]
    )

    b1 = np.array(
        [
            35.0 / 384.0,
            0.0,
            500.0 / 1113.0,
            125.0 / 192.0,
            -2187.0 / 6784.0,
            11.0 / 84.0,
            0.0,
        ]
    )
    b2 = np.array(
        [
            5179.0 / 57600.0,
            0.0,
            7571.0 / 16695.0,
            393.0 / 640.0,
            -92097.0 / 339200.0,
            187.0 / 2100.0,
            1.0 / 40.0,
        ]
    )

    truncation_error = float("inf")

    dt_elapsed = 0.0
    h = dt

    # Loop until we've gotten to our desired dt
    while dt_elapsed < dt:
        while truncation_error > max_error:
            # Only allow us to advance up to the dt remaining
            h = min(h, dt - dt_elapsed)

            k1 = f(x, u)
            k2 = f(x + h * (A[0, 0] * k1), u)
            k3 = f(x + h * (A[1, 0] * k1 + A[1, 1] * k2), u)
            k4 = f(x + h * (A[2, 0] * k1 + A[2, 1] * k2 + A[2, 2] * k3), u)
            k5 = f(
                x + h * (A[3, 0] * k1 + A[3, 1] * k2 + A[3, 2] * k3 + A[3, 3] * k4), u
            )
            k6 = f(
                x
                + h
                * (
                    A[4, 0] * k1
                    + A[4, 1] * k2
                    + A[4, 2] * k3
                    + A[4, 3] * k4
                    + A[4, 4] * k5
                ),
                u,
            )

            # Since the final row of A and the array b1 have the same coefficients
            # and k7 has no effect on newX, we can reuse the calculation.
            new_x = x + h * (
                A[5, 0] * k1
                + A[5, 1] * k2
                + A[5, 2] * k3
                + A[5, 3] * k4
                + A[5, 4] * k5
                + A[5, 5] * k6
            )
            k7 = f(new_x, u)

            truncation_error = np.linalg.norm(
                h
                * (
                    (b1[0] - b2[0]) * k1
                    + (b1[1] - b2[1]) * k2
                    + (b1[2] - b2[2]) * k3
                    + (b1[3] - b2[3]) * k4
                    + (b1[4] - b2[4]) * k5
                    + (b1[5] - b2[5]) * k6
                    + (b1[6] - b2[6]) * k7
                )
            )

            if truncation_error == 0.0:
                h = dt - dt_elapsed
            else:
                h *= 0.9 * math.pow(max_error / truncation_error, 1.0 / 5.0)

        dt_elapsed += h
        x = new_x

    return x
